#![allow(dead_code)]

mod util;
pub mod tag;

type Handle = u64;
pub use tag::Class;

pub fn init() -> Instance { Instance::new() }

fn mask_tag   (tag:Class)     -> Handle { (tag as u64) << 56 }
fn unmask_tag (handle:Handle) -> Class  { (handle >> 56) as Class }
fn handle_id  (handle:Handle) -> usize  { (handle & !mask_tag(0xff)) as usize }

struct Container<T> {
    pub class:usize,
    pub data:T,
}

pub struct Instance {
    scope:stdu::Trie<Handle>,
    type_tree:util::TypeTree,
    pool_boolean:stdu::Pool<bool>,
    pool_natural:stdu::Pool<u64>,
    pool_integer:stdu::Pool<i64>,
    pool_block:stdu::Pool<Vec<u8>>,
    pool_string:stdu::Pool<String>,
    pool_option:stdu::Pool<Container<Option<Handle>>>,
    pool_array:stdu::Pool<Container<Vec<Handle>>>,
    pool_list:stdu::Pool<Container<Vec<Handle>>>,
    pool_sparse:stdu::Pool<Container<stdu::Sparse<Handle>>>,
}
impl Instance {
    pub(crate) fn new() -> Self
    {
        Self {
            scope:stdu::Trie::<Handle>::new(),
            type_tree:util::TypeTree::new(),
            pool_boolean:stdu::Pool::<bool>::new(),
            pool_natural:stdu::Pool::<u64>::new(),
            pool_integer:stdu::Pool::<i64>::new(),
            pool_block:stdu::Pool::<Vec<u8>>::new(),
            pool_string:stdu::Pool::<String>::new(),
            pool_option:stdu::Pool::<Container<Option<Handle>>>::new(),
            pool_array:stdu::Pool::<Container<Vec<Handle>>>::new(),
            pool_list:stdu::Pool::<Container<Vec<Handle>>>::new(),
            pool_sparse:stdu::Pool::<Container<stdu::Sparse<Handle>>>::new(),
        }
    }


    /* DATA CONSTRUCTORS */
 
    pub fn boolean(&mut self, value:bool) -> Handle {
        mask_tag(tag::BOOLEAN) + self.pool_boolean.add(value) as Handle
    }

    pub fn natural(&mut self, value:u64) -> Handle {
        mask_tag(tag::NATURAL) + self.pool_natural.add(value) as Handle
    }

    pub fn integer(&mut self, value:i64) -> Handle {
        mask_tag(tag::INTEGER) + self.pool_integer.add(value) as Handle
    }

    pub fn block(&mut self, value:&Vec<u8>) -> Handle {
        mask_tag(tag::BLOCK) + self.pool_block.add(value.clone()) as Handle
    }

    pub fn string(&mut self, value:&str) -> Handle {
        mask_tag(tag::STRING) + self.pool_string.add(value.to_string()) as Handle
    }


    /* ENCODING/DECODING */

    fn encode_tags(tags:Vec<Class>) -> Vec<u8>
    {
        let mut encoded = Vec::<u8>::with_capacity(16);
        for tag in tags {
            encoded.append(&mut util::pack_natural(tag as u64));
        }
        return encoded;
    }

    fn encode_parts(&self, handle:Handle) -> Result<(Vec<Class>, Vec<u8>), ()>
    {
        let mut tags = Vec::<Class>::with_capacity(3);
        let mut encoded = Vec::<u8>::with_capacity(64);

        match unmask_tag(handle) {
            tag::NULL => {
                tags.push(tag::NULL);
            }
            tag::BOOLEAN => match self.pool_boolean.get(handle_id(handle)) {
                Some(value) => {
                    tags.push(tag::BOOLEAN + *value as Class);

                }
                None => { return Err(()); }
            }
            tag::NATURAL => match self.pool_natural.get(handle_id(handle)) {
                Some(value) => {
                    tags.push(tag::NATURAL);
                    encoded.append(&mut util::pack_natural(*value));
                }
                None => { return Err(()); }
            }
            tag::INTEGER => match self.pool_integer.get(handle_id(handle)) {
                Some(value) => {
                    tags.push(tag::INTEGER);
                    encoded.append(&mut util::pack_integer(*value));
                }
                None => { return Err(()); }
            }
            tag::BLOCK => match self.pool_block.get(handle_id(handle)) {
                Some(data) => {
                    tags.push(tag::BLOCK);
                    for b in data {
                        encoded.push(*b);
                    }
                }
                None => { return Err(()); }
            }
            tag::STRING => match self.pool_string.get(handle_id(handle)) {
                Some(data) => {
                    tags.push(tag::STRING);
                    let mut bytes = Vec::<u8>::from(data.as_bytes());
                    encoded.append(&mut util::pack_natural(bytes.len() as u64));
                    encoded.append(&mut bytes);
                }
                None => { return Err(()); }
            }
            tag::ARRAY => match self.pool_array.get(handle_id(handle)) {
                Some(_object) => {
                    tags.push(tag::ARRAY);
                    //tags.append(&mut object.class.clone());

                }
                None => { return Err(()); }
            }
            tag::LIST => match self.pool_list.get(handle_id(handle)) {
                Some(_object) => {
                    tags.push(tag::LIST);
                    //tags.append(&mut object.class.clone());

                }
                None => { return Err(()); }
            }
            tag::SPARSE => match self.pool_sparse.get(handle_id(handle)) {
                Some(_object) => {
                    tags.push(tag::SPARSE);
                    //tags.append(&mut object.class.clone());

                    // lookup table


                    // data table

                }
                None => { return Err(()); }
            }
            _ => { return Err(()); }
        }
        Err(())
    }

    pub fn encode_data(&self, handle:Handle) -> Result<Vec<u8>,()>
    {
        match self.encode_parts(handle) {
            Ok((_, encoded)) => Ok(encoded),
            Err(_) => Err(())
        }
    }

    pub fn encode(&self, handle:Handle) -> Result<Vec<u8>,()>
    {
        match self.encode_parts(handle) {
            Ok((tags, data)) => {
                let prefix = Self::encode_tags(tags);
                let mut encoded = Vec::<u8>::with_capacity(prefix.len() + data.len());
                for b in prefix { encoded.push(b); }
                for b in data { encoded.push(b); }

                Ok(encoded)
            }
            Err(_) => Err(())
        }
    }

    fn decode_tags(data:&Vec<u8>, index:&mut usize) -> Vec<Class>
    {
        let mut tags = Vec::<Class>::new();
        while *index < data.len() {
            let tag = util::unpack_natural(data, index) as Class;
            tags.push(tag);
            match tag {
                tag::ARRAY | tag::LIST | tag::SPARSE |tag::RECORD => { }
                _ => { break; }
            }
        }
        return tags;
    }

    pub fn decode(&mut self, data:&Vec<u8>) -> Result<Handle,()>
    {
        let mut index :usize = 0;
        let tags = Self::decode_tags(data, &mut index);

        for tag in tags {
            println!("{}", tag);
        }

        Err(())
    }
}
